from lib.database.entities.trust_trace import TrustTrace
from lib.database.entities.user_trust import UserTrust
from lib.trust.oracle import TrustOracle
from lib.database.entities.trust_adapter import TrustAdapter
from lib.database.entities.user import User
from typing import Dict, List
import logging
from datetime import datetime

LOGGER = logging.getLogger(__name__)


class TrustService:

    def __init__(self, adapter_repository, user_repository, user_trust_repository, trust_trace_repository):
        self._adapter_repository = adapter_repository
        self._user_repository = user_repository
        self._user_trust_repository = user_trust_repository
        self._trust_trace_repository = trust_trace_repository

    def updateSystem(self, use_case: str, table: str, properties: Dict):
        adapters: List[TrustAdapter] = [TrustAdapter.from_serializable_dict(row) for row in self._adapter_repository.all_for_use_case_and_table(
            use_case, table)]

        if len(adapters) == 0:
            LOGGER.warning(
                f"No Trust-Adapters found for Trace(use-case:{use_case}, table:{table}), aborting...")
            return

        for adapter in adapters:
            if adapter.user_source not in properties.keys() or adapter.volume_source not in properties.keys() or adapter.rating_source not in properties.keys():
                raise ValueError(
                    f"Trace does not contain needed fields for trust-computation. Required fields: {adapter.user_source}, {adapter.volume_source}. Present fields: {properties.keys()}")

            username: str = properties[adapter.user_source]
            volume: float = properties[adapter.volume_source]
            rating: float = properties[adapter.rating_source]

            # retrieve user (create one if it doesnt exist)
            user: User = self._user_repository.one_by_use_case_and_username(
                use_case, username)
            if user == None:
                user = User(use_case, username)
                self._user_repository.add(user)

            trace: TrustTrace = TrustTrace(
                use_case,
                adapter.context,
                user.name,
                adapter.table,
                volume,
                datetime.now(),
                is_trustworthy=(((rating - adapter.rating_lower) /
                                 (adapter.rating_upper - adapter.rating_lower)) * 2) - 1
            )

            # retrieve user_trust for the trace (create one if it doesnt exist)
            target_trust: UserTrust = None
            for user_trust in user.context_trust:
                if user_trust.context == adapter.context:
                    target_trust = user_trust
                    break
            if target_trust == None:
                target_trust = UserTrust(
                    use_case, adapter.context, user.name, 0, datetime(1, 1, 1, 0, 0))
                self._user_trust_repository.add(target_trust)
                user.context_trust.append(target_trust)

            oracle: TrustOracle = TrustOracle(
                adapter.conversion, adapter.certainty)
            delta_t = oracle.calculate_trust_single(trace)

            target_trust.t += delta_t
            target_trust.last_update = trace.timestamp

            self._trust_trace_repository.add(trace)
            self._user_trust_repository.update(target_trust)
